﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2008 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Core;
using System.Xml;
using System.Diagnostics;
using HeuristicLab.DataAnalysis;
using HeuristicLab.Operators;
using HeuristicLab.Random;
using HeuristicLab.Selection;
using HeuristicLab.Logging;
using HeuristicLab.Data;
using HeuristicLab.Operators.Programmable;
using HeuristicLab.Selection.OffspringSelection;

namespace HeuristicLab.GP.StructureIdentification {
  public class OffspringSelectionGP : StandardGP {

    private IntData maxEvaluatedSolutions = new IntData();
    public int MaxEvaluatedSolutions {
      get { return maxEvaluatedSolutions.Data; }
      set { maxEvaluatedSolutions.Data = value; }
    }

    private DoubleData selectionPressureLimit = new DoubleData();
    public double SelectionPressureLimit {
      get { return selectionPressureLimit.Data; }
      set { selectionPressureLimit.Data = value; }
    }

    private DoubleData comparisonFactor = new DoubleData();
    public double ComparisonFactor {
      get { return comparisonFactor.Data; }
      set { comparisonFactor.Data = value; }
    }

    private DoubleData successRatioLimit = new DoubleData();
    public double SuccessRatioLimit {
      get { return successRatioLimit.Data; }
      set { successRatioLimit.Data = value; }
    }

    public OffspringSelectionGP() : base() {
      PopulationSize = 1000;
      Parents = 20;
      MaxEvaluatedSolutions = 1000000;
      SelectionPressureLimit = 300;
      ComparisonFactor = 1.0;
      SuccessRatioLimit = 1.0;
    }

    internal override IOperator CreateGlobalInjector() {
      VariableInjector injector = (VariableInjector)base.CreateGlobalInjector();
      injector.RemoveVariable("TournamentSize");
      injector.RemoveVariable("MaxGenerations");
      injector.AddVariable(new HeuristicLab.Core.Variable("MaxEvaluatedSolutions", maxEvaluatedSolutions));
      injector.AddVariable(new HeuristicLab.Core.Variable("ComparisonFactor", comparisonFactor));
      injector.AddVariable(new HeuristicLab.Core.Variable("SelectionPressureLimit", selectionPressureLimit));
      injector.AddVariable(new HeuristicLab.Core.Variable("SuccessRatioLimit", successRatioLimit));
      return injector;
    }

    internal override IOperator CreateSelector() {
      CombinedOperator selector = new CombinedOperator();
      selector.Name = "Selector";
      SequentialProcessor seq = new SequentialProcessor();
      seq.Name = "Selector";
      EmptyOperator emptyOp = new EmptyOperator();
      ProportionalSelector femaleSelector = new ProportionalSelector();
      femaleSelector.GetVariableInfo("Selected").ActualName = "Parents";
      femaleSelector.GetVariableValue<BoolData>("CopySelected", null, false).Data = true;

      RandomSelector maleSelector = new RandomSelector();
      maleSelector.GetVariableInfo("Selected").ActualName = "Parents";
      maleSelector.GetVariableValue<BoolData>("CopySelected", null, false).Data = true;
      SequentialSubScopesProcessor seqSubScopesProc = new SequentialSubScopesProcessor();
      RightChildReducer rightChildReducer = new RightChildReducer();
      SubScopesMixer mixer = new SubScopesMixer();

      seqSubScopesProc.AddSubOperator(femaleSelector);
      seqSubScopesProc.AddSubOperator(emptyOp);

      seq.AddSubOperator(maleSelector);
      seq.AddSubOperator(seqSubScopesProc);
      seq.AddSubOperator(rightChildReducer);
      seq.AddSubOperator(mixer);

      selector.OperatorGraph.AddOperator(seq);
      selector.OperatorGraph.InitialOperator = seq;
      return selector;
    }

    internal override IOperator CreateMainLoop() {
      CombinedOperator main = new CombinedOperator();
      SequentialProcessor seq = new SequentialProcessor();
      SequentialProcessor offspringSelectionSeq = new SequentialProcessor();
      SequentialSubScopesProcessor subScopesProc = new SequentialSubScopesProcessor();
      EmptyOperator emptyOp = new EmptyOperator();
      OffspringSelector offspringSelector = new OffspringSelector();

      OperatorExtractor selector = new OperatorExtractor();
      selector.Name = "Selector (extr.)";
      selector.GetVariableInfo("Operator").ActualName = "Selector";

      IOperator childCreater = CreateChildCreater();
      childCreater.Name = "Create children";
      IOperator replacement = CreateReplacement();
      replacement.Name = "Replacement";
      IOperator solutionStorer = CreateBestSolutionProcessor();
      BestAverageWorstQualityCalculator qualityCalculator = new BestAverageWorstQualityCalculator();
      BestAverageWorstQualityCalculator validationQualityCalculator = new BestAverageWorstQualityCalculator();
      validationQualityCalculator.Name = "ValidationQualityCalculator";
      validationQualityCalculator.GetVariableInfo("BestQuality").ActualName = "BestValidationQuality";
      validationQualityCalculator.GetVariableInfo("AverageQuality").ActualName = "AverageValidationQuality";
      validationQualityCalculator.GetVariableInfo("WorstQuality").ActualName = "WorstValidationQuality";
      DataCollector collector = new DataCollector();
      ItemList<StringData> names = collector.GetVariable("VariableNames").GetValue<ItemList<StringData>>();
      names.Add(new StringData("BestQuality"));
      names.Add(new StringData("AverageQuality"));
      names.Add(new StringData("WorstQuality"));
      names.Add(new StringData("BestValidationQuality"));
      names.Add(new StringData("AverageValidationQuality"));
      names.Add(new StringData("WorstValidationQuality"));
      LinechartInjector lineChartInjector = new LinechartInjector();
      lineChartInjector.GetVariableInfo("Linechart").ActualName = "Quality Linechart";
      lineChartInjector.GetVariable("NumberOfLines").GetValue<IntData>().Data = 6;
      QualityLogger qualityLogger = new QualityLogger();
      QualityLogger validationQualityLogger = new QualityLogger();
      validationQualityCalculator.Name = "ValidationQualityLogger";
      validationQualityLogger.GetVariableInfo("Quality").ActualName = "ValidationQuality";
      validationQualityLogger.GetVariableInfo("QualityLog").ActualName = "ValidationQualityLog";
      Counter counter = new Counter();
      counter.GetVariableInfo("Value").ActualName = "Generations";
      LessThanComparator selPresComparator = new LessThanComparator();
      selPresComparator.GetVariableInfo("LeftSide").ActualName = "SelectionPressure";
      selPresComparator.GetVariableInfo("RightSide").ActualName = "SelectionPressureLimit";
      selPresComparator.GetVariableInfo("Result").ActualName = "SelectionPressureCondition";
      LessThanComparator generationsComparator = new LessThanComparator();
      generationsComparator.GetVariableInfo("LeftSide").ActualName = "EvaluatedSolutions";
      generationsComparator.GetVariableInfo("RightSide").ActualName = "MaxEvaluatedSolutions";
      generationsComparator.GetVariableInfo("Result").ActualName = "EvaluatedSolutionsCondition";
      ConditionalBranch selPresCondition = new ConditionalBranch();
      selPresCondition.GetVariableInfo("Condition").ActualName = "SelectionPressureCondition";
      ConditionalBranch generationsCondition = new ConditionalBranch();
      generationsCondition.GetVariableInfo("Condition").ActualName = "EvaluatedSolutionsCondition";

      subScopesProc.AddSubOperator(emptyOp);
      subScopesProc.AddSubOperator(childCreater);

      offspringSelector.AddSubOperator(offspringSelectionSeq);

      offspringSelectionSeq.AddSubOperator(selector);
      offspringSelectionSeq.AddSubOperator(subScopesProc);
      offspringSelectionSeq.AddSubOperator(offspringSelector);

      seq.AddSubOperator(offspringSelectionSeq);
      seq.AddSubOperator(replacement);
      seq.AddSubOperator(solutionStorer);
      seq.AddSubOperator(qualityCalculator);
      seq.AddSubOperator(validationQualityCalculator);
      seq.AddSubOperator(collector);
      seq.AddSubOperator(lineChartInjector);
      seq.AddSubOperator(qualityLogger);
      seq.AddSubOperator(validationQualityLogger);
      seq.AddSubOperator(counter);
      seq.AddSubOperator(selPresComparator);
      seq.AddSubOperator(generationsComparator);
      seq.AddSubOperator(selPresCondition);
      selPresCondition.AddSubOperator(generationsCondition);
      generationsCondition.AddSubOperator(seq);

      main.OperatorGraph.AddOperator(seq);
      main.OperatorGraph.InitialOperator = seq;
      return main;
    }

    public override IEditor CreateEditor() {
      return new OffspringSelectionGpEditor(this);
    }

    public override IView CreateView() {
      return new OffspringSelectionGpEditor(this);
    }

    public override object Clone(IDictionary<Guid, object> clonedObjects) {
      OffspringSelectionGP clone = (OffspringSelectionGP)base.Clone(clonedObjects);
      clone.SelectionPressureLimit = SelectionPressureLimit;
      clone.SuccessRatioLimit = SuccessRatioLimit;
      clone.ComparisonFactor = ComparisonFactor;
      return clone;
    }

    #region SetReferences Method
    private void SetReferences() {
      // SGA
      CombinedOperator co1 = (CombinedOperator)Engine.OperatorGraph.InitialOperator;
      // SequentialProcessor in SGA
      SequentialProcessor algorithm = (SequentialProcessor)co1.OperatorGraph.InitialOperator;
      // RandomInjector
      RandomInjector ri = (RandomInjector)algorithm.SubOperators[1];
      // VariableInjector
      VariableInjector vi = (VariableInjector)algorithm.SubOperators[2];
      maxEvaluatedSolutions = vi.GetVariable("MaxEvaluatedSolutions").GetValue<IntData>();
      selectionPressureLimit = vi.GetVariable("SelectionPressureLimit").GetValue<DoubleData>();
      successRatioLimit = vi.GetVariable("SuccessRatioLimit").GetValue<DoubleData>();
      comparisonFactor = vi.GetVariable("ComparisonFactor").GetValue<DoubleData>();
    }
    #endregion

    #region Persistence Methods
    public override XmlNode GetXmlNode(string name, XmlDocument document, IDictionary<Guid, IStorable> persistedObjects) {
      XmlNode node = base.GetXmlNode(name, document, persistedObjects);
      return node;
    }
    public override void Populate(XmlNode node, IDictionary<Guid, IStorable> restoredObjects) {
      base.Populate(node, restoredObjects);
      SetReferences();
    }
    #endregion
  }
}
